from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Iterable


class WildcardCollection(ABC):
    """
    A wildcard collection, i.e. a collection of strings.
    """

    @abstractmethod
    def get_values(self) -> Iterable[str]:
        """
        Get the contents of this collection.
        """
        raise NotImplementedError("...")



from __future__ import annotations

import logging

from dynamicprompts.commands import (
    Command,
    LiteralCommand,
    SequenceCommand,
    VariantCommand,
    WildcardCommand,
)
from dynamicprompts.commands.variable_commands import (
    VariableAccessCommand,
    VariableAssignmentCommand,
)
from dynamicprompts.sampling_context import SamplingContext
from dynamicprompts.types import StringGen
from dynamicprompts.utils import rotate_and_join

logger = logging.getLogger(__name__)


class Sampler:
    def generator_from_command(
        self,
        command: Command,
        context: SamplingContext,
    ) -> StringGen:
        # This is purposely not a dict lookup/getattr magic thing, to make
        # it easier for code completion etc. to see what's going on.
        if isinstance(command, LiteralCommand):
            return self._get_literal(command, context)
        if isinstance(command, SequenceCommand):
            return self._get_sequence(command, context)
        if isinstance(command, VariantCommand):
            return self._get_variant(command, context)
        if isinstance(command, WildcardCommand):
            return self._get_wildcard(command, context)
        if isinstance(command, VariableAssignmentCommand):
            raise NotImplementedError(
                "VariableAssignmentCommand should never be sampled",
            )
        if isinstance(command, VariableAccessCommand):
            return self._get_variable(command, context)
        return self._unsupported_command(command)

    def _unsupported_command(self, command: Command) -> StringGen:
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support {command.__class__.__name__}",
        )

    def _get_wildcard(
        self,
        command: WildcardCommand,
        context: SamplingContext,
    ) -> StringGen:
        return self._unsupported_command(command)

    def _get_variant(
        self,
        command: VariantCommand,
        context: SamplingContext,
    ) -> StringGen:
        return self._unsupported_command(command)

    def _get_sequence(
        self,
        command: SequenceCommand,
        context: SamplingContext,
    ) -> StringGen:
        tokens, context = context.process_variable_assignments(command.tokens)
        sub_generators = [context.generator_from_command(c) for c in tokens]

        while True:
            yield rotate_and_join(sub_generators, separator=command.separator)

    def _get_literal(
        self,
        command: LiteralCommand,
        context: SamplingContext,
    ) -> StringGen:
        while True:
            yield command.literal

    def _get_variable(
        self,
        command: VariableAccessCommand,
        context: SamplingContext,
    ) -> StringGen:
        variable = command.name
        command_to_sample = context.variables.get(variable, command.default)
        if not command_to_sample:
            if context.unknown_variable_value is None:
                raise KeyError(f"Variable {variable} is not defined in this context")
            elif isinstance(context.unknown_variable_value, str):
                command_to_sample = LiteralCommand(context.unknown_variable_value)
            else:
                command_to_sample = context.unknown_variable_value
        return context.for_sampling_variable(variable).generator_from_command(
            command_to_sample,
        )



from __future__ import annotations

import re

from dynamicprompts.utils import removeprefix, removesuffix


def clean_wildcard(wildcard: str, *, wildcard_wrap: str) -> str:
    """
    Clean, canonicalize, and validate a wildcard string (possibly wrapped with a `wrap`).

    :return: the cleaned wildcard
    """

    # remove wrapping
    wildcard = removeprefix(wildcard, wildcard_wrap)
    wildcard = removesuffix(wildcard, wildcard_wrap)

    # normalize path separators to forward slash
    wildcard = re.sub(r"[\\/]+", "/", wildcard)

    # remove leading and trailing path separators (likely a typo)
    wildcard = wildcard.strip("/")

    # validate
    if ".." in wildcard:
        raise ValueError(f"Wildcard can not contain '..': {wildcard}")
    if "\\" in wildcard:
        raise ValueError(f"Wildcard can not contain '\\': {wildcard}")
    return wildcard



from __future__ import annotations

import dataclasses
from typing import Any

from dynamicprompts.wildcards.collection.base import WildcardCollection


@dataclasses.dataclass(frozen=True)
class ListWildcardCollection(WildcardCollection):
    """
    A wildcard collection that's just a list of wildcards.

    The default implementation uses this to represent wildcards loaded from
    a structured pantry JSON/YAML file.
    """

    entries: list[str]

    # Implementation-specific hint for the source of the wildcards.
    source: Any = None

    def get_values(self) -> list[str]:
        return self.entries



from __future__ import annotations

import dataclasses

from dynamicprompts.commands import Command


@dataclasses.dataclass(frozen=True)
class VariableAssignmentCommand(Command):
    name: str
    value: Command
    immediate: bool
    sampling_method = None


@dataclasses.dataclass(frozen=True)
class VariableAccessCommand(Command):
    name: str
    default: Command | None = None
    sampling_method = None



from dynamicprompts.commands import VariantCommand, VariantOption, WildcardCommand
from dynamicprompts.parser.parse import parse
from dynamicprompts.sampling_context import SamplingContext


def wildcard_to_variant(
    command: WildcardCommand,
    *,
    context: SamplingContext,
    min_bound=1,
    max_bound=1,
    separator=",",
) -> VariantCommand:
    values = context.wildcard_manager.get_all_values(command.wildcard)
    min_bound = min(min_bound, len(values))
    max_bound = min(max_bound, len(values))

    variant_options = [
        VariantOption(parse(v, parser_config=context.parser_config)) for v in values
    ]

    wildcard_variant = VariantCommand(
        variant_options,
        min_bound,
        max_bound,
        separator,
        command.sampling_method,
    )
    return wildcard_variant



from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import Iterable

from dynamicprompts.wildcards.collection import WildcardCollection, WildcardTextFile
from dynamicprompts.wildcards.collection.structured import parse_structured_file
from dynamicprompts.wildcards.tree.tree import WildcardTree

log = logging.getLogger(__name__)


def _get_wildcard_file_name_from_path(path: Path, root_path: Path):
    if path == root_path:
        # Special case, because `with_suffix()` would fail
        return ""
    return str(path.relative_to(root_path).with_suffix("")).replace(os.sep, "/")


def _get_wildcard_collections_from_path(
    *,
    file_path: Path,
    root_path: Path,
) -> Iterable[tuple[str, WildcardCollection]]:
    if file_path.suffix == ".txt":
        name = _get_wildcard_file_name_from_path(file_path, root_path)
        yield (name, WildcardTextFile(file_path))
        return

    if file_path.suffix in (".json", ".yaml"):
        pantry_root_prefix = _get_wildcard_file_name_from_path(
            path=file_path.parent,
            root_path=root_path,
        )
        if pantry_root_prefix:
            # Suffix the / here, so we don't need to worry about pantries
            # in the root later
            pantry_root_prefix += "/"
        try:
            for name, wildcard_file in parse_structured_file(file_path):
                yield (f"{pantry_root_prefix}{name}", wildcard_file)
        except Exception as exc:  # pragma: no cover
            log.warning(
                "Unable to read structured wildcard file %s: %s",
                file_path,
                exc,
            )


def build_tree_from_path(root_path: Path) -> WildcardTree:
    """
    Walk a directory tree to build a WildcardTree.

    :param root_path: Root filesystem path.
    """
    path_to_file: dict[str, WildcardCollection] = {}

    for dirname, dirnames, filenames in os.walk(root_path, followlinks=True):
        dirnames[:] = [d for d in dirnames if not d.startswith(".")]
        for filename in filenames:
            path_to_file.update(
                _get_wildcard_collections_from_path(
                    file_path=Path(dirname) / filename,
                    root_path=root_path,
                ),
            )
    return WildcardTree.from_map(path_to_file)



from __future__ import annotations

import dataclasses
from fnmatch import fnmatch
from typing import Iterable

from dynamicprompts.wildcards.collection import WildcardCollection
from dynamicprompts.wildcards.tree.node import WildcardTreeNode


@dataclasses.dataclass(frozen=True)
class WildcardTree:
    root: WildcardTreeNode = dataclasses.field(default_factory=WildcardTreeNode)
    map: dict[str, WildcardCollection] = dataclasses.field(default_factory=dict)

    def __post_init__(self):
        if not self.map:
            self.update_file_map()

    def update_file_map(self) -> None:
        self.map.clear()
        self.map.update(self.root.walk_items())

    @classmethod
    def from_map(cls, map: dict[str, WildcardCollection]) -> WildcardTree:
        """
        Build a WildcardTree from a map of paths to WildcardCollections.
        """
        root = WildcardTreeNode("")
        for path, collection in map.items():
            parts = path.split("/")
            node = root
            for part in parts[:-1]:
                if part not in node.child_nodes:
                    node.child_nodes[part] = WildcardTreeNode(node.qualify_name(part))
                node = node.child_nodes[part]
            node.collections[parts[-1]] = collection
        return cls(root, map=map.copy())

    def get_collection(self, path: str) -> WildcardCollection:
        return self.map[path]

    def match_collections(self, pattern: str) -> Iterable[WildcardCollection]:
        for name, coll in self.map.items():
            if fnmatch(name, pattern):
                yield coll

    def get_collection_names(self) -> Iterable[str]:
        return self.map.keys()



from dynamicprompts.wildcards.collection.base import WildcardCollection
from dynamicprompts.wildcards.collection.text_file import WildcardTextFile

__all__ = [
    "WildcardCollection",
    "WildcardTextFile",
]



from __future__ import annotations

from pathlib import Path

from dynamicprompts.constants import DEFAULT_ENCODING
from dynamicprompts.utils import is_empty_line
from dynamicprompts.wildcards.collection.base import WildcardCollection


class WildcardTextFile(WildcardCollection):
    """
    A wildcard collection that is stored in a text file.
    """

    def __init__(
        self,
        path: Path,
        encoding: str = DEFAULT_ENCODING,
    ) -> None:
        self._path = path
        self._encoding = encoding
        self._cache: set[str] | None = None

    def __repr__(self) -> str:
        return f"<TxtWildcardFile: {self._path}>"

    def get_values(self) -> set[str]:
        if self._cache is not None:
            return self._cache

        with self._path.open(encoding=self._encoding, errors="ignore") as f:
            lines = [line.strip() for line in f if not is_empty_line(line)]
            self._cache = set(lines)
            return self._cache

    def read_text(self) -> str:
        """
        Read the file's raw contents as a string.
        """
        return self._path.read_text(encoding=self._encoding, errors="ignore")

    def write_text(self, contents: str) -> None:
        """
        Rewrite the file's contents with the given string.
        """
        self._path.write_text(contents, encoding=self._encoding)
        self._cache = None



from __future__ import annotations

import dataclasses
from typing import Iterable

from dynamicprompts.wildcards.collection import WildcardCollection


@dataclasses.dataclass(frozen=True)
class WildcardTreeNode:
    path: str = ""
    child_nodes: dict[str, WildcardTreeNode] = dataclasses.field(default_factory=dict)
    collections: dict[str, WildcardCollection] = dataclasses.field(default_factory=dict)

    def get_items(self) -> Iterable[tuple[str, WildcardCollection]]:
        """
        Get the collection items (pair of full path -> collection) of this node.
        """
        for name, file in self.collections.items():
            yield (self.qualify_name(name), file)

    def qualify_name(self, name: str) -> str:
        """
        Qualify a collection name with the path of this node.
        """
        return f"{self.path}/{name}".strip("/")

    def get_full_names(self) -> Iterable[str]:
        """
        Get the full paths of the files in this node.
        """
        for name in self.collections:
            yield self.qualify_name(name)

    def walk_items(self) -> Iterable[tuple[str, WildcardCollection]]:
        """
        Walk the items (pair of full path -> collection) of this node and its children.
        """
        yield from self.get_items()
        for child in self.child_nodes.values():
            yield from child.walk_items()

    def walk_full_names(self) -> Iterable[str]:
        """
        Walk the full paths of the collections in this node and its children.
        """
        yield from self.get_full_names()
        for child in self.child_nodes.values():
            yield from child.walk_full_names()



from __future__ import annotations

from dynamicprompts.wildcards.tree.build import build_tree_from_path
from dynamicprompts.wildcards.tree.node import WildcardTreeNode
from dynamicprompts.wildcards.tree.tree import WildcardTree

__all__ = [
    "WildcardTree",
    "WildcardTreeNode",
    "build_tree_from_path",
]



from __future__ import annotations

import json
import logging
import warnings
from pathlib import Path
from typing import Iterable

from dynamicprompts.constants import DEFAULT_ENCODING
from dynamicprompts.wildcards.collection.base import WildcardCollection
from dynamicprompts.wildcards.collection.list import ListWildcardCollection

log = logging.getLogger(__name__)


def _parse_structured_file_dict(
    data: dict,
    *,
    file_path: Path,
    prefix: tuple[str, ...] = (),
) -> Iterable[tuple[str, WildcardCollection]]:
    """
    Parse a single dict level in a structured file (JSON or YAML) and yield the wildcard collections.
    """
    for name, value in data.items():
        if not isinstance(name, str):
            continue
        prefix_and_name = (*prefix, name)
        name = "/".join(prefix_and_name)
        if isinstance(value, list) and all(isinstance(x, str) for x in value):
            yield (
                name,
                ListWildcardCollection(entries=value, source=(file_path, name)),
            )
        elif isinstance(value, dict):
            yield from _parse_structured_file_dict(
                value,
                file_path=file_path,
                prefix=prefix_and_name,
            )
        else:
            log.warning(
                "Wildcard file %s has unsupported value for key %s",
                file_path,
                name,
            )


def parse_structured_file(file_path: Path) -> Iterable[tuple[str, WildcardCollection]]:
    """
    Parse a structured file (JSON or YAML) and yield the wildcard collections.
    """
    content = file_path.read_text(encoding=DEFAULT_ENCODING)

    if file_path.suffix == ".yaml":
        try:
            import yaml

            data = yaml.safe_load(content)
        except ImportError:  # pragma: no cover
            warnings.warn("YAML support is not available, skipping YAML wildcard file")
            return ()
    elif file_path.suffix == ".json":
        data = json.loads(content)
    else:  # pragma: no cover
        raise ValueError(f"Unexpected file extension: {file_path.suffix}")

    if isinstance(data, dict):
        return _parse_structured_file_dict(data, file_path=file_path)
    if isinstance(data, list):
        # See if the data is a flat list of strings
        valid_entries = [e for e in data if isinstance(e, str)]
        if valid_entries:
            # Only return a collection if it's not empty.
            collection = ListWildcardCollection(
                entries=valid_entries,
                source=file_path,
            )
            return [(file_path.with_suffix("").name, collection)]
    return ()



from __future__ import annotations

import logging
from pathlib import Path
from typing import Iterable

from dynamicprompts.parser.config import default_parser_config
from dynamicprompts.wildcards.collection import WildcardCollection
from dynamicprompts.wildcards.tree import (
    WildcardTree,
    build_tree_from_path,
)
from dynamicprompts.wildcards.utils import clean_wildcard

logger = logging.getLogger(__name__)


class WildcardManager:
    def __init__(
        self,
        path: Path | str | None = None,
        wildcard_wrap=default_parser_config.wildcard_wrap,
    ) -> None:
        self._path: Path | None = Path(path) if path else None
        self._wildcard_wrap = wildcard_wrap
        self._tree: WildcardTree | None = None

    @property
    def path(self) -> Path | None:
        """
        The root path of the wildcard manager, if set.
        """
        return self._path

    @property
    def wildcard_wrap(self) -> str:
        """
        The string that is used as the prefix and suffix of a wildcard
        the default is "__" (two underscores)
        """
        return self._wildcard_wrap

    def to_wildcard(self, name: str) -> str:
        """
        Wrap `name` in the wildcard wrap string if it is not already wrapped.
        """
        ww = self._wildcard_wrap
        if not name.startswith(ww):
            name = f"{ww}{name}"
        if not name.endswith(ww):
            name = f"{name}{ww}"
        return name

    def is_wildcard(self, text: str) -> bool:
        """
        Check if `text` is a wildcard reference (i.e. starts and ends with the wildcard wrap string)
        """
        return text.startswith(self.wildcard_wrap) and text.endswith(self.wildcard_wrap)

    @property
    def tree(self) -> WildcardTree:
        """
        Get the wildcard tree.

        If the tree has not been built yet, it will be built from the configured path.
        """
        if self._tree is None:
            self._tree = (
                build_tree_from_path(self._path)
                if self._path and self._directory_exists()
                else WildcardTree()
            )
        return self._tree

    def clear_cache(self) -> None:
        """
        Clear the cache of the wildcard manager.
        """
        self._tree = None

    def _directory_exists(self) -> bool:
        return bool(self._path and self._path.is_dir())

    def match_collections(self, wildcard: str) -> Iterable[WildcardCollection]:
        """
        Find `WildcardCollection` objects that match the given glob pattern.
        """
        if not self._path:
            return []
        try:
            wildcard = clean_wildcard(wildcard, wildcard_wrap=self._wildcard_wrap)
        except ValueError:
            logger.warning(f"Invalid wildcard: {wildcard}", exc_info=True)
            return []
        return self.tree.match_collections(wildcard)

    def get_file(self, name: str) -> WildcardCollection:
        """
        Get a single wildcard collection by full name; raise an exception if there is no match.
        """
        name = clean_wildcard(name, wildcard_wrap=self._wildcard_wrap)
        return self.tree.get_collection(name)

    def get_collection_names(self) -> set[str]:
        """
        Get the names of all known wildcard collections.
        """
        return set(self.tree.get_collection_names())

    def get_all_values(self, wildcard: str) -> list[str]:
        """
        Get all wildcard values matching the given wildcard pattern.
        """
        values: set[str] = set()
        for f in self.match_collections(wildcard):
            values.update(f.get_values())
        return sorted(values)
